import typing as t


class BaseRegistryError(Exception):
    pass


class ChildMissingLocator(BaseRegistryError):
    pass


class ChildMissingLocatorValue(BaseRegistryError):
    pass


class ChildNotFound(BaseRegistryError):
    pass


class RegistriesMixin:

    __registries: dict[str, dict[t.Any, type]] = {}

    __registry_type__: str  # default: class name
    __registry_locator__: str

    @classmethod
    def __get_type(cls):
        return getattr(cls, "__registry_type__", cls.__name__)

    @classmethod
    def __get_loc(cls):
        loc = getattr(cls, "__registry_locator__", None)
        if loc is None:
            raise ChildMissingLocator(cls)
        return loc

    @classmethod
    def __get_loc_val(cls):
        value = getattr(cls, cls.__get_loc(), None)
        if value is None:
            raise ChildMissingLocatorValue(cls)
        return value

    def __init_subclass__(cls, **kwargs: t.Any):
        super().__init_subclass__(**kwargs)
        mixin = RegistriesMixin

        if mixin in cls.__bases__:
            cls.__registry_type__ = cls.__get_type()  # enrich
            cls.__get_loc()  # validation
        else:
            store = mixin.__registries.setdefault(cls.__get_type(), {})
            store[cls.__get_loc_val()] = cls

    @classmethod
    def get_child_for(cls, locator_value: t.Any) -> type:
        store = cls.__registries[cls.__registry_type__]
        if locator_value not in store:
            raise ChildNotFound(locator_value)
        return store[locator_value]



import itertools
import random

import requests
from requests import compat


class Endpoints:

    def __init__(self, endpoints):
        random.shuffle(endpoints)
        self._endpoints = itertools.cycle(endpoints)
        self._endpoint = next(self._endpoints)

    def shift(self):
        self._endpoint = next(self._endpoints)

    def make_url(self, path) -> str:
        return compat.urljoin(self._endpoint, path)


class HttpClient(requests.Session):

    def __init__(self, base, timeout=1, auth=None, verify=True):
        if isinstance(base, str):
            base = Endpoints([base])
        self._base = base
        self._timeout = timeout
        super().__init__()
        self.auth = auth
        self.verify = verify

    def prepare_request(self, request: requests.Request,
                        ) -> requests.PreparedRequest:
        url = request.url.lower()
        for prefix in self.adapters.keys():
            if url.startswith(prefix.lower()):
                break
        else:
            request.url = self._base.make_url(request.url)

        return super().prepare_request(request)

    def send(self, request, timeout=None, **kwargs):
        if timeout is None:
            timeout = self._timeout

        try:
            return super().send(request, timeout=timeout, **kwargs)
        except Exception:
            self._base.shift()
            raise



from miska import registries


def test_registry():
    class Registry(registries.RegistriesMixin):
        __registry_locator__ = "MY_TYPE"

    class MyDriver(Registry):
        MY_TYPE = "mine!"

    result = Registry.get_child_for("mine!")

    assert result == MyDriver
